#!/bin/python3
# SPDX-FileCopyrightInfo: Copyright © DUNE-FUFEM Project contributors, see file AUTHORS.md
# SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception or LGPL-3.0-or-later


import os
import yaml
import sys
import getopt
import subprocess


def usage():
  print('This program should be run inside of a dune modules source directory.')
  print('It traverses the path upwards until a dune.module file is found to')
  print('identify the module root directory. In this directory it parses the')
  print('.gitlab-ci.yml file for a code-spelling-check job and then executes')
  print('this jobs script command. In contrast to running codespell directly')
  print('this respects the same rules and ignore list as the CI job.')
  print('')
  print('Options are:')
  print(' --help, -h:  Show this help text')

try:
  opts, args = getopt.getopt(sys.argv[1:], "h", ["help"])
except getopt.GetoptError:
  usage()
  sys.exit(2)

for opt, arg in opts:
  if opt in ("-h", "--help"):
    usage()
    sys.exit()

# Traverse directory tree upwards until we find a dune.module file
modulePath = os.path.realpath(os.path.curdir)
while not os.path.exists(modulePath+"/dune.module"):
  parentPath = os.path.realpath(modulePath+"/..")
  if parentPath==modulePath:
    print("Error: No dune.module file found in ancestor directories")
    sys.exit(1)
  modulePath = parentPath
print("Module base directory is              : "+modulePath)

# Parse codespell command from .gitlab-ci.yml
try:
  with open(modulePath+"/.gitlab-ci.yml", "r") as f:
    tree = yaml.safe_load(f)
    command = tree["code-spelling-check"]["script"][0]
    command += " -S .git"
except:
  print("Error: Could not parse codespell command from .gitlab-ci.yml")
  sys.exit(1)
print("Codespell command from .gitlab-ci.yml : "+command)

if len(sys.argv[1:]) > 0:
  print("Additional options from command line  : "+(" ".join(sys.argv[1:])))

# Execute codespell command
try:
  print("Running codespell in module directory")
  subprocess.run(command.split(' ')+sys.argv[1:], cwd=modulePath)
except:
  print("Error: Failed running codespell")
  sys.exit(1)
