@page poisson-pq2-cpp Poisson equation
<!--
SPDX-FileCopyrightInfo: Copyright © DUNE-FUFEM Project contributors, see file AUTHORS.md
SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception or LGPL-3.0-or-later
-->

@dontinclude{lineno} poisson-pq2.cc

The following explains how to solve the Poisson equation
using [dune-fufem][]. The full example @ref source-poisson-pq2-cpp
can be found in the `examples/` subdirectory.

### Setup and initialization


After including the required headers the `main()` function
first creates a simple logger that allows to write log
messages to `std::cout`. The logger will prefix log messages
with the time passed since the program was started and since
the last log message using the given format string.
Then the first two command line arguments - if present - are
parsed. They allow to set the number of threads used
in parallel algorithms and the number of grid refinement steps.
Afterward MPI is initialized, passing the command line arguments.

@skip LABEL_START_MAIN
@until LABEL_END_SETUP

### Define problem data

The example solves the Poisson equation with Dirichlet
boundary conditions. The problem data consisting of
the right-hand-side function and the Dirichlet values
are defined using lambda expressions.

@until LABEL_END_PROBLEMDATA

### Create grid and finite element basis

Instead of creating a uniform grid using
a factory or reading a grid from a file, the example
manually creates a mixed grid of type `UGGrid<2>`
using a `GridFactory`. The grid implementation `UGGrid<2>`
supports unstructured grid in two dimensions. After inserting nine
uniformly placed vertices on the unit square, the method
inserts two quadrilateral elements in the lower half and
four triangular elements in the upper half of the domain.
Finally, the call to `factory.createGrid()` returns a `std::unique_ptr`
referring to the newly creates grid. For simplicity the example
also stores a reference to the grid.
The coarse grid is then refined globally and copy of the
leaf grid view is stored.

@note
A `GridView` is a view to a subset of the grid's elements, vertices, ...
that should be stored by value and can be copied cheaply.
Grids in [dune][] are in general hierarchical and composed by elements
on several levels. The discretization usually lives on the set of
most refined elements that is denote the leaf `GridView` in dune.

@until LABEL_END_CREATEGRID

Dune-fufem supports thread-parallel assembly based
on a coloring of the grid view. To use this feature we
need to compute such a coloring.

@until LABEL_END_COLORING

As a next step the program creates a global finite element
basis on the `GridView`. The helper function
`lagrange<2>()` declares that we want to use Lagrange
finite elements of order two, while the `makeBasis()`
function creates the basis of this type on the given
`GridView`. Both functions are from the namespace
`Dune::Functions::BasisFactory` which is included
for simplicity.

@until LABEL_END_MAKEBASIS

### Create matrix, vector, and constraints data structures

Since the basis` indices are plain integers we next define
types for flat vectors and matrices of reals numbers
and create a solution and a right-hand-side vector
and a stiffness matrix.

@until LABEL_END_MATRIXVECTORTYPES

To handle the Dirichlet boundary conditions we the
create an object for storing affine constraints
with respect to the basis and then fill the object
by computing the actual constraints values on the
boundary from the Dirichlet value function.

@until LABEL_END_CONSTRAINTS

### Assemble the variational problem

Inspired by the UFL language of the [Fenics project][fenics] project,
dune-fufem allows to define local assemblers for linear and bilinear
forms using simple expressions from the @ref Dune::Fufem::Forms namespace.
After including the namespace, we first obtain expressions for trial and
test functions associated to the basis. The only difference between
`trialFunction()` and `testFunction()` is that the resulting objects
later refer to column and row indices, respectively.
In order to use the right-hand-side lambda functions within a form
expression we first wrap it into a `GridView`-functions and then
declare this as coefficient function, i.e. a function not depending
on test and trial functions.
Now we can define the bilinear and linear form in a straight forward way.

@until LABEL_END_LOCALASSEMBLER

The defined bilinear form and linear form objects
satisfy the interfaces for local matrix and vector assemblers
and can thus directly be passed to a global assembler
for assembling the actual matrix and right-hand-side
vector. In order to use thread-parallel assembly, we
additionally pass the grid coloring and thread count
to the assembler. Once the system is assembled, we
can constrain it to the affine subspace satisfying
the constraints that we defined earlier.

@until LABEL_END_GLOBALASSEMBLER

### Solve the algebraic problem

Before solving the linear system, we have to initialize
the solution vector. While we could directly use the
corresponding resize method and assignment operator,
we here demonstrate how the `istlVectorBackend()`
from [dune-functions][] can be used, since this approach
would also work for basis using multi-indices and
corresponding nested vector containers.

@until LABEL_END_INITSOLUTION

To use the matrix in the solvers from the [dune-istl][] module
we first wrap it into a suitable operator interface.
Then we create an AMG preconditioner and pass the operator
and preconditioner to a CG solver. The `apply()` method
of the latter solves the linear system up to the given
tolerance and writes some solver statistics to the
respective object that needs to be passed additionally.

@until LABEL_END_APPLYSOLVER

### Write solution to VTK

Finally we can write the solution to a VTK file understood by
the [ParaView][] visualization program. To this end we need
to create a finite element function by combining the basis
with the coefficient vector. Here this is obtained using
the expression `u|sol` which binds the linear operator `u`
(i.e. the identity on the finite element space) to the coefficient
vector `sol`.
We create a writer object capable of writing unstructured grids
and tell it to write polynomial data of order 2 and pass the
function to the writer, specifying the name and data type
of the resulting field in the output.

@until LABEL_END_WRITEVTK

### Error handling

In case an exception is thrown within `main()` it is followed by a
`catch` block that will print the error message and exit.

@until LABEL_END_MAIN


[dune]: https://dune-project.org
[dune-istl]: https://gitlab.dune-project.org/core/dune-istl
[dune-fufem]: https://gitlab.dune-project.org/fufem/dune-fufem
[dune-functions]: https://gitlab.dune-project.org/staging/dune-functions
[dune-vtk]: https://gitlab.dune-project.org/extensions/dune-vtk
[fenics]: https://fenicsproject.org
[ParaView]: https://www.paraview.org


<div class="section_buttons">
| Previous      |                         Next |
|:--------------|-----------------------------:|
| @ref examples |  @ref poisson-pq2-py         |
</div>



@page source-poisson-pq2-cpp poisson-pq2.cc

This is the raw source code of the poisson-pq2.cc example.
There is also a [commented version](@ref poisson-pq2-cpp)
of this example in the @ref examples section.

@include{lineno} poisson-pq2.cc
