// SPDX-FileCopyrightText: Copyright © DUNE-FUFEM Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#include <config.h>

#include <vector>
#include <thread>
#include <mutex>
#include <sstream>
#include <regex>

#include <dune/common/rangeutilities.hh>
#include <dune/common/test/testsuite.hh>

#include <dune/fufem/logger.hh>


template<class Duration=void>
auto checkLogger(std::string format, std::string prefixRegex)
{
  auto testSuite = Dune::TestSuite(std::string("Logger check with format '")+format+"'");

  auto callMakeLogger = [](auto&&... args) {
    if constexpr(std::is_same_v<Duration,void>)
      return Dune::Fufem::makeLogger(std::forward<decltype(args)>(args)...);
    else
      return Dune::Fufem::makeLogger<Duration>(std::forward<decltype(args)>(args)...);
  };

  auto callMakeSynchronizedLogger = [](auto&&... args) {
    if constexpr(std::is_same_v<Duration,void>)
      return Dune::Fufem::makeSynchronizedLogger(std::forward<decltype(args)>(args)...);
    else
      return Dune::Fufem::makeSynchronizedLogger<Duration>(std::forward<decltype(args)>(args)...);
  };
  
  // Checking serial logger
  {
    auto sstream = std::ostringstream();
    auto logger = callMakeLogger(sstream, format);
    logger("foo");
    logger("bar");
    auto regex = prefixRegex+"foo\n"+prefixRegex+"bar\n";
    testSuite.check(std::regex_match(sstream.str(), std::regex(regex)))
      << "Non-matching string generated by logger with format '" << format <<"'";
  }

  // Checking synchronized logger
  {
    auto sstream = std::ostringstream();
    auto mutex = std::mutex();
    auto logger = callMakeSynchronizedLogger(sstream, format, mutex);
    auto threads = std::vector<std::thread>();
    for(auto i: Dune::range(3))
      threads.emplace_back([&]() {
        for([[maybe_unused]] auto i: Dune::range(3))
        {
          using namespace std::chrono_literals;
          std::this_thread::sleep_for(10ms);
          logger("foo");
        }
      });
    for(auto& thread: threads)
      thread.join();
    auto regex = std::string();
    for([[maybe_unused]] auto i: Dune::range(9))
      regex = regex+prefixRegex+"foo\n";
    testSuite.check(std::regex_match(sstream.str(), std::regex(regex)))
      << "Non-matching string generated by synchronized logger with format '" << format <<"'";
  }

  return testSuite;
}



int main(int argc, char** argv)
{

  auto testSuite = Dune::TestSuite();

  
  // Check with default duration

  testSuite.subTest(checkLogger("[% 8.3f] [% 8.3f] ", "\\[[0-9 .]{8}\\] \\[[0-9 .]{8}\\] "));

  // Check with explicitly given duration
  testSuite.subTest(checkLogger<std::chrono::duration<double>>("[% 8.3f] [% 8.3f] ", "\\[[0-9 .]{8}\\] \\[[0-9 .]{8}\\] "));

  testSuite.subTest(checkLogger<std::chrono::milliseconds>("% 5d:% 5d:", "[0-9 ]{5}\\:[0-9 ]{5}\\:"));

  return testSuite.exit();
}
