// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// SPDX-FileCopyrightText: Copyright © DUNE-FUFEM Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#include <config.h>

#include <vector>
#include <cmath>
#include <numbers>

#include <dune/common/exceptions.hh>
#include <dune/common/stringutility.hh>
#include <dune/common/rangeutilities.hh>

#include <dune/geometry/type.hh>

#include <dune/grid/uggrid.hh>
#include <dune/grid/utility/structuredgridfactory.hh>

#include <dune/istl/bvector.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/solvers.hh>

#include <dune/functions/backends/istlvectorbackend.hh>
#include <dune/functions/functionspacebases/defaultglobalbasis.hh>
#include <dune/functions/functionspacebases/nedelecbasis.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/vtk/vtkwriter.hh>
#include <dune/vtk/datacollectors/discontinuouslagrangedatacollector.hh>

#include <dune/fufem/logger.hh>
#include <dune/fufem/parallel/elementcoloring.hh>
#include <dune/fufem/forms/forms.hh>
#include <dune/fufem/constraints/affineconstraints.hh>
#include <dune/fufem/constraints/boundaryconstraints.hh>

#include "utilities.hh"

int main (int argc, char *argv[]) try
{

  //! Create a simple logger
  auto log = Dune::Fufem::makeLogger(std::cout, "[% 8.3f] [% 8.3f] ");

  //! Process command line arguments
  std::size_t threadCount = 4;
  if (argc>1)
    threadCount = std::stoul(std::string(argv[1]));

  std::size_t refinements = 4;
  if (argc>2)
    refinements = std::stoul(std::string(argv[2]));

  log("Command line processed");

  Dune::MPIHelper::instance(argc, argv);

  log("MPI initialized");

  using Range = Dune::FieldVector<double,2>;
  auto rightHandSide = [] (const auto& x) {
    if (std::abs(x[0]-x[1]) < .2)
      return Range({1,1});
    else
      return Range({0,0});
  };
  auto dirichletValues = [](const auto& x){ return Range(0); };

  using Grid = Dune::UGGrid<2>;
  auto gridPtr = Dune::StructuredGridFactory<Grid>::createCubeGrid({{0,0}}, {{1,1}}, {{1,1}});
  auto& grid = *gridPtr;

  log("Grid created");

  grid.globalRefine(refinements);

  log(Dune::formatString("Grid refined %d times", refinements));

  auto gridView = grid.leafGridView();

  auto gridViewPartition = Dune::Fufem::coloredGridViewPartition(gridView);

  log("Grid coloring computed");

  using namespace Dune::Functions::BasisFactory;
  auto basis = makeBasis(gridView, nedelec<1,1>());

  log(Dune::formatString("Basis created with dimension %d", basis.dimension()));

  using Vector = Dune::BlockVector<double>;
  using Matrix = Dune::BCRSMatrix<double>;

  auto sol = Vector();
  auto rhs = Vector();
  auto stiffnessMatrix = Matrix();

  auto constraints = Dune::Fufem::makeAffineConstraints(basis);

  Dune::Fufem::computeBoundaryConstraints(constraints, basis, dirichletValues);

  log("Boundary constraints computed");

  using namespace Dune::Fufem::Forms;

  auto u = trialFunction(basis, NonAffineFamily());
  auto v = testFunction(basis, NonAffineFamily());
  auto f = Coefficient(Dune::Functions::makeGridViewFunction(rightHandSide, gridView));

  auto a = integrate(dot(curl(u), curl(v)) + dot(u,v));
  auto l = integrate(dot(f,v));

  auto globalAssembler = GlobalAssembler(basis, basis, 0, std::cref(gridViewPartition), threadCount);

  log("Assembler set up");

  globalAssembler.assembleOperator(stiffnessMatrix, a);

  log("Matrix assembled");

  globalAssembler.assembleFunctional(rhs, l);

  log("RHS assembled");

  constraints.constrainLinearSystem(stiffnessMatrix, rhs);

  log("Boundary condition incorporated into system");

  Dune::Functions::istlVectorBackend(sol).resize(basis);
  Dune::Functions::istlVectorBackend(sol) = 0;

  auto matrixOperator = Dune::MatrixAdapter<Matrix,Vector,Vector>(stiffnessMatrix);

  auto preconditioner = makeAMGPreconditioner<Vector>(stiffnessMatrix);

  log("Preconditioner created");

  auto solver = Dune::CGSolver<Vector>(
                  matrixOperator,
                  preconditioner,
                  1e-4,   // desired residual reduction factor
                  1000,   // maximum number of iterations
                  2       // verbosity of the solver
                );

  log("Solver created");

  auto solverStatistics = Dune::InverseOperatorResult();

  solver.apply(sol, rhs, solverStatistics);

  log("Linear system solved");

  auto vtkWriter = Dune::Vtk::UnstructuredGridWriter(Dune::Vtk::DiscontinuousLagrangeDataCollector(basis.gridView(), 2));
  vtkWriter.addPointData(u|sol, Dune::VTK::FieldInfo("sol", Dune::VTK::FieldInfo::Type::vector, 2));
  vtkWriter.write("curlcurl");

  log("Solution written to vtk file");

}
catch (Dune::Exception& e)
{
  std::cout << e.what() << std::endl;
  return 1;
}
