// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// SPDX-FileCopyrightText: Copyright © DUNE-FUFEM Project contributors, see file AUTHORS.md
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception OR LGPL-3.0-or-later

#include <config.h>

#include <vector>
#include <cmath>
#include <numbers>

#include <dune/common/exceptions.hh>
#include <dune/common/stringutility.hh>
#include <dune/common/rangeutilities.hh>

#include <dune/geometry/type.hh>

#include <dune/grid/uggrid.hh>

#include <dune/istl/bvector.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/solvers.hh>

#include <dune/functions/backends/istlvectorbackend.hh>
#include <dune/functions/functionspacebases/defaultglobalbasis.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/vtk/vtkwriter.hh>
#include <dune/vtk/datacollectors/lagrangedatacollector.hh>

#include <dune/fufem/logger.hh>
#include <dune/fufem/parallel/elementcoloring.hh>
#include <dune/fufem/forms/forms.hh>
#include <dune/fufem/constraints/affineconstraints.hh>
#include <dune/fufem/constraints/boundaryconstraints.hh>

#include "utilities.hh"

//! LABEL_START_MAIN
int main (int argc, char *argv[]) try
{

  //! Create a simple logger
  auto log = Dune::Fufem::makeLogger(std::cout, "[% 8.3f] [% 8.3f] ");

  //! Process command line arguments
  std::size_t threadCount = 4;
  if (argc>1)
    threadCount = std::stoul(std::string(argv[1]));

  std::size_t refinements = 4;
  if (argc>2)
    refinements = std::stoul(std::string(argv[2]));

  log("Command line processed");

  Dune::MPIHelper::instance(argc, argv);

  log("MPI initialized");
//! LABEL_END_SETUP

  auto rightHandSide = [] (const auto& x) { return 10;};
  auto dirichletValues = [](const auto& x){ return std::sin(2*std::numbers::pi*x[0]); };
//! LABEL_END_PROBLEMDATA

  using Grid = Dune::UGGrid<2>;

  auto factory = Dune::GridFactory<Grid>();
  for(unsigned int k : Dune::range(9))
    factory.insertVertex({0.5*(k%3), 0.5*(k/3)});
  factory.insertElement(Dune::GeometryTypes::cube(2), {0, 1, 3, 4});
  factory.insertElement(Dune::GeometryTypes::cube(2), {1, 2, 4, 5});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {3, 4, 6});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {4, 7, 6});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {4, 5, 7});
  factory.insertElement(Dune::GeometryTypes::simplex(2), {5, 8, 7});

  auto gridPtr = std::unique_ptr(factory.createGrid());
  auto& grid = *gridPtr;

  log("Grid created");

  grid.globalRefine(refinements);

  log(Dune::formatString("Grid refined %d times", refinements));

  auto gridView = grid.leafGridView();
//! LABEL_END_CREATEGRID

  auto gridViewPartition = Dune::Fufem::coloredGridViewPartition(gridView);

  log("Grid coloring computed");
//! LABEL_END_COLORING

  using namespace Dune::Functions::BasisFactory;
  auto basis = makeBasis(gridView, lagrange<5>());

  log(Dune::formatString("Basis created with dimension %d", basis.dimension()));
//! LABEL_END_MAKEBASIS

  using Vector = Dune::BlockVector<double>;
  using Matrix = Dune::BCRSMatrix<double>;

  auto sol = Vector();
  auto rhs = Vector();
  auto stiffnessMatrix = Matrix();
//! LABEL_END_MATRIXVECTORTYPES

  auto constraints = Dune::Fufem::makeAffineConstraints(basis);

  Dune::Fufem::computeBoundaryConstraints(constraints, basis, dirichletValues);

  log("Boundary constraints computed");
//! LABEL_END_CONSTRAINTS

  using namespace Dune::Fufem::Forms;

  auto u = trialFunction(basis);
  auto v = testFunction(basis);
  auto f = Coefficient(Dune::Functions::makeGridViewFunction(rightHandSide, gridView));

  auto a = integrate(dot(grad(u), grad(v)));
  auto l = integrate(f*v);
//! LABEL_END_LOCALASSEMBLER

  auto globalAssembler = GlobalAssembler(basis, basis, 0, std::cref(gridViewPartition), threadCount);

  log("Assembler set up");

  globalAssembler.assembleOperator(stiffnessMatrix, a);

  log("Matrix assembled");

  globalAssembler.assembleFunctional(rhs, l);

  log("RHS assembled");

  constraints.constrainLinearSystem(stiffnessMatrix, rhs);

  log("Boundary condition incorporated into system");
//! LABEL_END_GLOBALASSEMBLER

  Dune::Functions::istlVectorBackend(sol).resize(basis);
  Dune::Functions::istlVectorBackend(sol) = 0;
//! LABEL_END_INITSOLUTION

  auto matrixOperator = Dune::MatrixAdapter<Matrix,Vector,Vector>(stiffnessMatrix);

  auto preconditioner = makeAMGPreconditioner<Vector>(stiffnessMatrix);

  log("Preconditioner created");

  auto solver = Dune::CGSolver<Vector>(
                  matrixOperator,
                  preconditioner,
                  1e-4,   // desired residual reduction factor
                  1000,   // maximum number of iterations
                  2       // verbosity of the solver
                );

  log("Solver created");

  auto solverStatistics = Dune::InverseOperatorResult();

  solver.apply(sol, rhs, solverStatistics);

  log("Linear system solved");
//! LABEL_END_APPLYSOLVER

  auto vtkWriter = Dune::Vtk::UnstructuredGridWriter(Dune::Vtk::LagrangeDataCollector(basis.gridView(), 2));
  vtkWriter.addPointData(u|sol, Dune::VTK::FieldInfo("sol", Dune::VTK::FieldInfo::Type::scalar, 1));
  vtkWriter.write("poisson-pq2");

  log("Solution written to vtk file");
//! LABEL_END_WRITEVTK

}
catch (Dune::Exception& e)
{
  std::cout << e.what() << std::endl;
  return 1;
}
//! LABEL_END_MAIN
